/*
 * Matrices.java
 *
 * Creado el 9 de marzo de 2006, 18:18
 */


package auxiliar;

/**
 *
 * @author Vctor Jimnez, Antonio Pallars,
 * Mara Jos Cano Vicente, Juan Sebastin Daz Garca,
 * Bernardo Gonzlez Merino
 *
 * MTODOS NUMRICOS
 * LICENCIATURA DE MATEMTICAS
 * UNIVERSIDAD DE MURCIA
 */

public class Matrices {
        
    static double precision = 5E-16;
    
    public static boolean ifCuadrada(double[][] A) {
        int n = A.length;
        boolean respuesta = true;
        for(int i = 0; i < n; i++) {
            if(A[i].length != n) {
                respuesta = false;
                break;
            }
        }
        return respuesta;
    }
    
    public static boolean ifMatriz(double[][] A) {
        int n = A.length;
        int m = A[0].length;
        boolean respuesta = true;
        
        for(int i = 0; i < n; i++) {
            if(A[i].length != m) {
                respuesta = false;
                break;
            }
        }
        return respuesta;
    }
    
    public static double[][] suma(double[][] A, double[][] B)
            throws ErrorMetodosLineales {
        int n;
        int m;
        double[][] AmasB;
        
        if (!ifMatriz(A)) {
            System.out.println("Error: matriz izquierda mal definida");
            throw new ErrorMetodosLineales();
        }
        if (!ifMatriz(B)) {
            System.out.println("Error: matriz derecha mal definida");
            throw new ErrorMetodosLineales();
        }
        n = B.length;
        m = B[0].length;
        if ((n != A.length) || (m != A[0].length)) {
            System.out.println("Error: dimensiones incompatibles");
            throw new ErrorMetodosLineales();
        }
        
        AmasB = new double[n][m];
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < m; j++) {
                AmasB[i][j] = A[i][j] + B[i][j];
            }
        }
        return AmasB;
    }
    
    public static double[] suma(double[] u, double[] v)
            throws ErrorMetodosLineales {
        double[][] A = new double [1][u.length];
        double[][] B = new double [1][v.length];
        
        A[0] = u;
        B[0] = v;
        return suma(A, B)[0];
    }
    
    public static double[][] resta(double[][] A, double[][] B)
            throws ErrorMetodosLineales {
        double[][] menosB = producto(-1, B);
        
        return suma(A, menosB);
    }
    
    public static double[] resta(double[] u, double[] v)
            throws ErrorMetodosLineales {
        double[] menosv = producto(-1, v);
        
        return suma(u, menosv);
    }
    
    public static double[][] producto(double[][] A, double[][] B)
            throws ErrorMetodosLineales {
        int n;
        int m;
        int p;
        double[][] AB;
        
        if (!ifMatriz(A)) {
            System.out.println("Error: matriz izquierda mal definida");
            throw new ErrorMetodosLineales();
        }
        if (!ifMatriz(B)) {
            System.out.println("Error: matriz derecha mal definida");
            throw new ErrorMetodosLineales();
        }
        n = A.length;
        m = B.length;
        p = B[0].length;
        if (m != A[0].length) {
            System.out.println("Error: dimensiones incompatibles");
            throw new ErrorMetodosLineales();
        }
        
        AB = new double[n][p];
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < p; j++) {
                AB[i][j] = 0;
                for (int k = 0; k < m; k++) {
                    AB[i][j] = AB[i][j] + A[i][k] * B[k][j];
                }
            }
        }
        return AB;
    }
    
    public static double[][] producto(double c, double[][] A)
            throws ErrorMetodosLineales {
        int n;
        int m;
        double[][] cA;
        
        if (!ifMatriz(A)) {
            System.out.println("Error: matriz mal definida");
            throw new ErrorMetodosLineales();
        }
        
        n = A.length;
        m = A[0].length;
        cA = new double[n][m];
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < m; j++) {
                cA[i][j] = c * A[i][j];
            }
        }
        return cA;
    }
    
    public static double[] producto(double c, double[] u)
            throws ErrorMetodosLineales {
        double[][] A = new double[1][u.length];
        
        A[0] = u;
        return producto(c, A)[0];
    }
    
    public static double producto(double[] u, double[] v)
            throws ErrorMetodosLineales {
        double[][] A = new double[1][u.length];
        double[][] B = new double[1][v.length];
        double[][] Bt = new double[v.length][1];
        
        A[0] = u;
        B[0] = v;
        Bt = traspuesta(B);
        
        return producto(A, Bt)[0][0];
    }
    
    public static double[] producto(double[][] A, double[] v)
            throws ErrorMetodosLineales {
        double[][] B = new double[1][v.length];
        double[][] Bt = new double[v.length][1];
        
        B[0] = v;
        Bt = traspuesta(B);
        return traspuesta(producto(A, Bt))[0];
    }
    
    public static double[][] traspuesta(double[][] A)
            throws ErrorMetodosLineales {
        int n;
        int m;
        double[][] At;
        
        if (!ifMatriz(A)) {
            System.out.println("Error: matriz mal definida");
            throw new ErrorMetodosLineales();
        }
        
        n = A.length;
        m = A[0].length;
        At = new double[m][n];
        for (int j = 0; j < m; j++) {
            for (int i = 0; i < n; i++) {
                At[j][i] = A[i][j];
            }
        }
        return At;
    }
    
    public static void escribe(double[][] A) {
        int n;
        String cadena;
        
        n = A.length;
        for (int i = 0; i < n; i++) {
            cadena = "(" + A[i][0];
            for (int j = 1; j < A[i].length; j++) {
                cadena = cadena + ", " + A[i][j];
            }
            cadena = cadena + ")";
            System.out.println(cadena);
        }
    }
    
    public static void escribe(double[] u) {
        double[][] A = new double[1][u.length];
        
        A[0] = u;
        escribe(A);
    }
    
    public static void escribe(int[][] A) {
        int n;
        String cadena;
        
        n = A.length;
        for (int i = 0; i < n; i++) {
            cadena = "(" + A[i][0];
            for (int j = 1; j < A[i].length; j++) {
                cadena = cadena + ", " + A[i][j];
            }
            cadena = cadena + ")";
            System.out.println(cadena);
        }
    }
    
    public static void escribe(int[] u) {
        int[][] A = new int[1][u.length];
        
        A[0] = u;
        escribe(A);
    }
    
    public static String toString(double[][] A) {
        int n;
        String cadena = "";
        
        n = A.length;
        for (int i = 0; i < n; i++) {
            cadena = cadena + "(" + A[i][0];
            for (int j = 1; j < A[i].length; j++) {
                cadena = cadena + ", " + A[i][j];
            }
            if (i < n - 1) {
                cadena = cadena + ")\n";
            } else {
                cadena = cadena + ")";
            }
        }
        return cadena;
    }
    
    public static String toString(double[] u) {
        double[][] A = new double[1][u.length];
        
        A[0] = u;
        return toString(A);
    }
    
    public static String toString(int[][] A) {
        int n;
        String cadena = "";
        
        n = A.length;
        for (int i = 0; i < n; i++) {
            cadena = cadena + "(" + A[i][0];
            for (int j = 1; j < A[i].length; j++) {
                cadena = cadena + ", " + A[i][j];
            }
            if (i < n - 1) {
                cadena = cadena + ")\n";
            } else {
                cadena = cadena + ")";
            }
        }
        return cadena;
    }
    
    public static String toString(int[] u) {
        int[][] A = new int[1][u.length];
        
        A[0] = u;
        return toString(A);
    }
    
    public static double norma(double[] u) {
        double norm = 0;
        
        for (int i = 0; i < u.length; i++) {
            norm = norm + u[i] * u[i];
        }
        return Math.sqrt(norm);
    }
    
    public static double norma1(double[][] A) throws ErrorMetodosLineales {
        int n = A.length;
        double norma;
        
        if(!ifCuadrada(A)) {
            System.out.println("Error: matriz no cuadrada");
            throw new ErrorMetodosLineales();
        }
        
        norma = 0;
        for(int j = 0; j < n; j++) {
            double suma = 0;
            for(int i = 0; i < n; i++) {
                suma = suma + Math.abs(A[i][j]);
            }
            if(norma < suma) {
                norma = suma;
            }
        }
        return norma;
    }

    public static double aproxRadioEspectral(double[][] A)
            throws ErrorMetodosLineales {
        double radio;
        double radioAprox = 0;
        int n = A.length;
        int k;
        double[][] IterA;
        
        if(!ifCuadrada(A)) {
            System.out.println("Error: matriz no cuadrada");
            throw new ErrorMetodosLineales();
        }       
        
        IterA = new double[n][n];
        for (int i = 0; i < n; i++) {
            for(int j = 0; j < n; j++) {
                IterA[i][j] = A[i][j];
            }
        }
        radio = Matrices.norma1(IterA);
        k = 1;
        while ( (radio != Double.NaN) && (radio != Double.POSITIVE_INFINITY)
                && (k <= 1000) )  {
            radioAprox = radio;
            System.out.println("||M^{" + k + "}||^{1/" + k + "} = " 
                    + radioAprox);
            k = k + 1;
            IterA = Matrices.producto(A, IterA);
            radio = Math.pow(Matrices.norma1(IterA), 1. / k);
        }
       
        return radioAprox;
    }
    
    public static double determinanteMenores(double[][] A)
            throws ErrorMetodosLineales {
        int n;
        int signo;
        double det;
        double[][] Aux;
        
        if(!ifCuadrada(A)) {
            System.out.println("Error: matriz no cuadrada");
            throw new ErrorMetodosLineales();
        }
        
        n = A.length;
        if (n == 1) {
            return A[0][0];
        } else {
            Aux = new double[n - 1][n - 1];
            det = 0.;
            signo = 1;
            for (int k = 0; k < n; k++) {
                for (int i = 0; i < k; i++) {
                    for (int j = 0; j < n - 1; j++) {
                        Aux[i][j] = A[i][j + 1];
                    }
                }
                for (int i = k; i < n - 1; i++) {
                    for (int j = 0; j < n - 1; j++) {
                        Aux[i][j] = A[i + 1][j + 1];
                    }
                }
                det = det + signo * A[k][0] * determinanteMenores(Aux);
                signo = - signo;
            }
            return det;
        }
        
    }
    
    public static double[] reglaCramer(double[][] A, double[] b)
            throws ErrorMetodosLineales {        
        int n = A.length;
        double[][] M;
        double[] x; 
        double denominador;
        
        if(!ifCuadrada(A)) {
            System.out.println("Error: matriz no cuadrada");
            throw new ErrorMetodosLineales();
        }
        if (b.length != n) {
            System.out.println("Error: dimensiones no compatibles");
            throw new ErrorMetodosLineales();
        }
        
        denominador = determinanteMenores(A);
        if (Math.abs(denominador) < precision) {
            System.out.println("Error: matriz singular");
            throw new ErrorMetodosLineales();
        }
        
        M = new double[n][n];
        x = new double[n];
        
        for (int k = 0; k < n; k ++) {
            
            /* Calculamos el coeficiente x_k del vector solucin */
            for (int i = 0; i < n; i++) {
                for(int j = 0; j < n; j++) {
                    if (j != k) {
                        M[i][j] = A[i][j];
                    } else {
                        M[i][j] = b[i];
                    }                    
                }
            }
            x[k] = determinanteMenores(M) / denominador;
        }           
        return x;
    }
    
    public static double[][] sen(double[][] A, int iter) 
            throws ErrorMetodosLineales {
        int n = A.length;
        double[][] B = new double[n][n];
        double[][] senA = new double[n][n];
        
        if (!ifCuadrada(A)) {
            System.out.println("Error: matriz no cuadrada");
            throw new ErrorMetodosLineales();
        }
      
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                B[i][j] = A[i][j];
                senA[i][j] = A[i][j];
            }
        }
        for (int k = 1; k <= iter; k++) {
            B = producto(-1. / (2 * k * (2 * k + 1)), producto(A, producto(A, B)));            
            senA = suma(senA, B);      
        }        
        return senA;
    }
    
    public static double[][] cos(double[][] A, int iter) 
            throws ErrorMetodosLineales {
        int n = A.length;
        double[][] B = new double[n][n];
        double[][] cosA = new double[n][n];
        
        if (!ifCuadrada(A)) {
            System.out.println("Error: matriz no cuadrada");
            throw new ErrorMetodosLineales();
        }
        
        for (int i = 0; i < n; i++) {
            B[i][i] = 1;
            cosA[i][i] = 1;
        }
        for (int k = 1; k <= iter; k++) {
            B = producto(-1. / ((2* k - 1) * 2 * k), producto(A, producto(A, B)));
            cosA = suma(cosA, B);
        }
        
        return cosA;
    }
    
    public static double[] gaussParcial(double[][] A, double[] b)
            throws ErrorMetodosLineales {
        int n = A.length;
        int m;
        int p;
        double mul;
        double[][] B; // matriz de los cambios
        double [] v; // vector de los cambios
        int[] s; // vector puntero
        double[] x; // vector donde escribir las soluciones
        
        if(!ifCuadrada(A)) {
            System.out.println("Error: matriz no cuadrada");
            throw new ErrorMetodosLineales();
        }
        if (b.length != n) {
            System.out.println("Error: dimensiones no compatibles");
            throw new ErrorMetodosLineales();
        }
        
        B = new double[n][n];
        v = new double[n];
        s = new int[n];
        x = new double[n];
        
        /* Condiciones iniciales */
        for (int i = 0; i < n; i++) {
            for(int j = 0; j < n; j++) {
                B[i][j] = A[i][j];
            }
        }
        for (int i = 0; i < n; i++) {
            v[i] = b[i];
        }
        for (int i = 0; i < n; i++) {
            s[i] = i;
        }
        
        /* Vamos a hacer las etapas k = 0, 1,..., n-1 */
        for (int k = 0; k < n; k++) {
            
            /* 1. Eleccin del pivote parcial */
            p = k; // iniciamos la bsqueda en la fila k
            for (int i = k + 1; i < n; i++) {
                if((Math.abs(B[s[i]][k])) > (Math.abs(B[s[p]][k]))) {
                    p = i; // apuntamos a la fila donde est el puntero
                }
            }
            if (Math.abs(B[s[p]][k]) < precision) {
                System.out.println("Error: matriz singular");
                System.out.println(toString(B));
                throw new ErrorMetodosLineales();
            }
            
            /* 2. Intercambio virtual de las filas pues cambiamos slo el puntero */
            m = s[k];
            s[k] = s[p];
            s[p] = m;
            
            /* 3. Eliminacin */
            for (int i = k + 1; i < n; i++) {
                mul = B[s[i]][k] / B[s[k]][k];
                B[s[i]][k] = 0; // asignamos ceros porque es el resultado de
                // las operaciones y as nos las ahorramos
                for(int j = k + 1; j < n; j++) {
                    B[s[i]][j] = B[s[i]][j] - mul * B[s[k]][j];
                }
                v[s[i]] = v[s[i]] - mul * v[s[k]];
            }
            
        } // ya tenemos la matriz de forma triangular superior
        
        /* Ahora resolvemos el sistema por el metodo ascendente */
        for (int k = n - 1; k >= 0; k--) {
            x[k] = v[s[k]];
            for (int j = k + 1; j < n; j++) {
                x[k] = x[k] - B[s[k]][j] * x[j];
            }
            x[k] = x[k] / B[s[k]][k];
        }
        return x;
    }
        
    /**
     * Excepcin ErrorMetodosLineales para ser devuelta por
     * los mtodos de la clase
     */
    public static class  ErrorMetodosLineales extends Exception {
        ErrorMetodosLineales() {
            super("Error: mtodos lineales");
        }
    }
}
